package seng302.Controllers;


import javafx.application.Platform;
import javafx.collections.ObservableList;
import javafx.fxml.FXML;
import javafx.scene.chart.LineChart;
import javafx.scene.chart.NumberAxis;
import javafx.scene.chart.XYChart;
import javafx.geometry.Pos;
import javafx.scene.control.*;
import javafx.scene.layout.GridPane;
import javafx.scene.paint.Color;
import javafx.scene.layout.Pane;
import javafx.scene.layout.StackPane;
import javafx.scene.paint.Color;
import javafx.scene.paint.Paint;
import seng302.Mock.StreamedRace;
import seng302.Model.*;
import seng302.VisualiserInput;

import java.net.URL;
import java.util.*;
import java.util.ArrayList;
import java.util.Locale;
import java.util.ResourceBundle;

/**
 * Created by fwy13 on 15/03/2017.
 */
public class RaceController extends Controller {
    @FXML GridPane canvasBase;

    //user saved data for annotation display
    private ArrayList<Boolean> presetAnno;
    private ArrayList<Boat> startBoats;
    private Integer sparkLineNumber = 0;
    private ResizableRaceCanvas raceMap;
    private ResizableRaceMap raceBoundaries;
    private ArrayList<String> colours;
    private Map<Integer, String> boatColours = new HashMap<>();
    private int legNum;
    private RaceClock raceClock;
    @FXML Pane arrow;
    @FXML SplitPane race;
    @FXML StackPane arrowPane;
    @FXML CheckBox showFPS;
    @FXML CheckBox showBoatPath;
    @FXML CheckBox showAnnotations;
    @FXML Label timer;
    @FXML Label FPS;
    @FXML Label timeZone;
    @FXML CheckBox showName;
    @FXML CheckBox showAbbrev;
    @FXML CheckBox showSpeed;
    @FXML CheckBox showTime;
    @FXML Button saveAnno;
    @FXML Button showSetAnno;
    @FXML TableView<Boat> boatInfoTable;
    @FXML TableColumn<Boat, String> boatPlacingColumn;
    @FXML TableColumn<Boat, String> boatTeamColumn;
    @FXML TableColumn<Boat, String> boatMarkColumn;
    @FXML TableColumn<Boat, String> boatSpeedColumn;
    @FXML LineChart<Number, Number> sparklineChart;
    @FXML NumberAxis xAxis;
    @FXML NumberAxis yAxis;

    /**
     * Updates the ResizableRaceCanvas (raceMap) with most recent data
     *
     * @param boats boats that are to be displayed in the race
     * @param boatMarkers Markers for boats
     * @see ResizableRaceCanvas
     */
    public void updateMap(ObservableList<Boat> boats, ObservableList<Marker> boatMarkers) {
        raceMap.setBoats(boats);
        raceMap.setBoatMarkers(boatMarkers);
        raceMap.update();
        raceBoundaries.draw();
        //stop if the visualiser is no longer running
    }

    /**
     * Updates the array listened by the TableView (boatInfoTable) that displays the boat information.
     *
     * @param race Race to listen to.
     */
    public void setInfoTable(StreamedRace race) {
        //boatInfoTable.getItems().clear();
        ObservableList<Boat> startingBoats = race.getStartingBoats();
        boatInfoTable.setItems(race.getStartingBoats());
        boatTeamColumn.setCellValueFactory(cellData -> cellData.getValue().getName());
        boatSpeedColumn.setCellValueFactory(cellData -> cellData.getValue().getVelocityProp());
        boatMarkColumn.setCellValueFactory(cellData -> cellData.getValue().getCurrentLegName());
        boatPlacingColumn.setCellValueFactory(cellData -> cellData.getValue().positionProperty());
    }


    @Override
    public void initialize(URL location, ResourceBundle resources) {
        //listener for fps
        startBoats = new ArrayList<>();
        showFPS.selectedProperty().addListener((ov, old_val, new_val) -> {
            if (showFPS.isSelected()) {
                FPS.setVisible(true);
            } else {
                FPS.setVisible(false);
            }
        });
    }

    /**
     * Creates and sets initial display for Sparkline for race positions.
     * A data series for each boat in the race is added.
     * Position numbers are displayed.
     *
     * @param boats boats to display on the sparkline
     */
    public void createSparkLine(ObservableList<Boat> boats){
        // NOTE: Y axis is in negatives to display correct positions

        makeColours();
        startBoats.addAll(boats);
        mapBoatColours();

        // all boats start in 'last' place
        for (int i=0; i<startBoats.size(); i++){
            XYChart.Series<Number, Number> series = new XYChart.Series();
            series.getData().add(new XYChart.Data(0, -startBoats.size()));
            series.getData().add(new XYChart.Data(0, -startBoats.size()));
            sparklineChart.getData().add(series);
            sparklineChart.getData().get(i).getNode().setStyle("-fx-stroke: " +
                    ""+boatColours.get(startBoats.get(i).getSourceID())+";");
        }

        sparklineChart.setCreateSymbols(false);

        // set x axis details
        xAxis.setAutoRanging(false);
        xAxis.setTickMarkVisible(false);
        xAxis.setTickLabelsVisible(false);
        xAxis.setMinorTickVisible(false);
        xAxis.setUpperBound(startBoats.size()*legNum);
        xAxis.setTickUnit(startBoats.size()*legNum);


        // set y axis details
        yAxis.setLowerBound(-(startBoats.size()+1));
        yAxis.setUpperBound(0);
        yAxis.setAutoRanging(false);
        yAxis.setLabel("Position in Race");
        yAxis.setTickUnit(1);
        yAxis.setTickMarkVisible(false);
        yAxis.setMinorTickVisible(false);
        // hide minus number from displaying on axis
        yAxis.setTickLabelFormatter(new NumberAxis.DefaultFormatter(yAxis) {
            @Override
            public String toString(Number value) {
                if ((Double)value == 0.0
                        || (Double)value < -startBoats.size()){
                    return "";
                }
                else {
                    return String.format("%7.0f", -value.doubleValue());
                }
            }
        });
    }

    /**
     * Initializes and runs the race, based on the user's chosen scale factor
     * Currently uses an example racecourse
     *
     * @param visualiserInput input from network
     * @param raceClock The RaceClock to use for the race's countdown/elapsed duration + timezone.
     */
    public void startRace(VisualiserInput visualiserInput, RaceClock raceClock) {

        //newRace.initialiseBoats();
        legNum = visualiserInput.getCourse().getLegs().size()-1;

        makeArrow();

        raceMap = new ResizableRaceCanvas(visualiserInput.getCourse());
        raceMap.setMouseTransparent(true);
        raceMap.widthProperty().bind(canvasBase.widthProperty());
        raceMap.heightProperty().bind(canvasBase.heightProperty());
        //raceMap.setBoats(newRace.getStartingBoats());
        raceMap.draw();
        raceMap.setVisible(true);
        raceMap.setArrow(arrow.getChildren().get(0));

        canvasBase.getChildren().add(0, raceMap);

        raceBoundaries = new ResizableRaceMap(visualiserInput.getCourse());
        raceBoundaries.setMouseTransparent(true);
        raceBoundaries.widthProperty().bind(canvasBase.widthProperty());
        raceBoundaries.heightProperty().bind(canvasBase.heightProperty());
        raceBoundaries.draw();
        raceBoundaries.setVisible(true);

        canvasBase.getChildren().add(0, raceBoundaries);

        race.setVisible(true);

        //Initialize save annotation array, fps listener, and annotation listeners
        timeZone.setText(raceClock.getTimeZone());

        //RaceClock.duration isn't necessarily being changed in the javaFX thread, so we need to runlater the update.
        raceClock.durationProperty().addListener((observable, oldValue, newValue) -> {
            Platform.runLater(() -> {
                timer.setText(newValue);
            });
        });

        this.raceClock = raceClock;
        raceMap.setRaceClock(raceClock);

        StreamedRace newRace = new StreamedRace(visualiserInput, this);
        initializeFPS();
        initializeAnnotations();

        new Thread((newRace)).start();
    }

    public void finishRace(ObservableList<Boat> boats){
        race.setVisible(false);
        parent.enterFinish(boats);
    }


    /**
     * Set the value for the race clock label
     *
     * @param time time that the label will be updated to
     */
    public void setTimer(String time) {
        //timer.setText(time);
    }

    /**
     * Set the value for the fps label
     *
     * @param fps fps that the label will be updated to
     */
    public void setFrames(String fps) {
        FPS.setText((fps));
    }

    /**
     * Set up FPS display at bottom of screen
     */
    private void initializeFPS() {
        showFPS.setVisible(true);
        showFPS.selectedProperty().addListener((ov, old_val, new_val) -> {
            if (showFPS.isSelected()) {
                FPS.setVisible(true);
            } else {
                FPS.setVisible(false);
            }
        });
    }

    /**
     * Updates the sparkline to display current boat positions.
     * New points are plotted to represent each boat when required.
     *
     * @param boatsInRace used for current boat positions.
     */
    public void updateSparkline(ObservableList<Boat> boatsInRace){
        int placingVal = boatsInRace.size();
        sparkLineNumber++;

        for (int i = boatsInRace.size() - 1; i >= 0; i--){
            for (int j = startBoats.size() - 1; j >= 0; j--){
                if (boatsInRace.get(i)==startBoats.get(j)){

                    // when a boat is on its first leg
                    if (boatsInRace.get(i).getCurrentLeg().getLegNumber()==0){
                        // adjust boats latest point on X axis
                        sparklineChart.getData().get(j).getData().get(1)
                                .setXValue(sparkLineNumber);
                    }

                    // when a boat first enters its second leg
                    else if (boatsInRace.get(i).getCurrentLeg().getLegNumber
                            ()==1 && sparklineChart.getData().get(j).getData
                            ().size()==2){
                        // adjust boats position from start mark
                        sparklineChart.getData().get(j).getData().get(1)
                                .setYValue(-placingVal);
                        sparklineChart.getData().get(j).getData().get(1)
                                .setXValue(sparkLineNumber);
                        sparklineChart.getData().get(j).getData().add(new XYChart.Data<>
                                (sparkLineNumber, -placingVal));
                    }

                    // plot new point for boats current position
                    else {
                        sparklineChart.getData().get(j).getData().add
                                (new XYChart.Data<>(sparkLineNumber, -placingVal));
                    }
                    placingVal-=1;
                }
            }
        }
//        xAxis.setUpperBound(sparkLineNumber);
//        xAxis.setTickUnit(sparkLineNumber);
    }

    private void makeColours() {
        colours = new ArrayList<>(Arrays.asList(
                colourToHex(Color.BLUEVIOLET),
                colourToHex(Color.BLACK),
                colourToHex(Color.RED),
                colourToHex(Color.ORANGE),
                colourToHex(Color.DARKOLIVEGREEN),
                colourToHex(Color.LIMEGREEN),
                colourToHex(Color.PURPLE),
                colourToHex(Color.DARKGRAY),
                colourToHex(Color.YELLOW)
        ));
    }

    private void mapBoatColours() {
        int currentColour = 0;
        for (Boat boat : startBoats) {
            if (!boatColours.containsKey(boat.getSourceID())) {
                boatColours.put(boat.getSourceID(), colours.get(currentColour));
            }
            currentColour = (currentColour + 1) % colours.size();
        }
    }

    /**
     * Set up boat annotations
     */
    private void initializeAnnotations() {
        presetAnno = new ArrayList<>();
        //listener for annotation
        showAnnotations.selectedProperty().addListener((ov, old_val, new_val) -> {
            raceMap.toggleAnnotations();
            raceMap.update();
        });
        //listener for show name in annotation
        showName.selectedProperty().addListener((ov, old_val, new_val) -> {
            raceMap.toggleAnnoName();
            raceMap.update();
        });
        //listener for show abbreviation for annotation
        showAbbrev.selectedProperty().addListener((ov, old_val, new_val) -> {
            raceMap.toggleAnnoAbbrev();
            raceMap.update();
        });

        //listener for show boat path for annotation
        showBoatPath.selectedProperty().addListener((ov, old_val, new_val) -> {
            raceMap.toggleBoatPath();
            raceMap.update();
        });
        //listener to show speed for annotation
        showSpeed.selectedProperty().addListener((ov, old_val, new_val) -> {
            raceMap.toggleAnnoSpeed();
            raceMap.update();
        });
        showTime.selectedProperty().addListener((ov, old_val, new_val) -> {
            raceMap.toggleAnnoTime();
            raceMap.update();
        });
        //listener to save currently selected annotation
        saveAnno.setOnAction(event -> {
            presetAnno.clear();
            presetAnno.add(showName.isSelected());
            presetAnno.add(showAbbrev.isSelected());
            presetAnno.add(showSpeed.isSelected());
            presetAnno.add(showBoatPath.isSelected());
        });
        //listener to show saved annotation
        showSetAnno.setOnAction(event -> {
            if (presetAnno.size() > 0) {
                showName.setSelected(presetAnno.get(0));
                showAbbrev.setSelected(presetAnno.get(1));
                showSpeed.setSelected(presetAnno.get(2));
                showBoatPath.setSelected(presetAnno.get(3));
                raceMap.update();
            }
        });
    }

    private String colourToHex(Color color) {
        return String.format( "#%02X%02X%02X",
                (int)( color.getRed() * 255 ),
                (int)( color.getGreen() * 255 ),
                (int)( color.getBlue() * 255 ) );
    }

    private void makeArrow() {
        arrowPane.getChildren().add(arrow);
    }

    public RaceClock getRaceClock() {
        return raceClock;
    }
}
